@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@
@ dispatch.s
@ thread scheduler and dispatcher
@
@ Created by DarkPhantom.
@
@ Modified by samel on 28/02/08.
@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
.GLOBL current_thread
.GLOBL ready_queue
.GLOBL suspended_queue
.GLOBL dispatcher
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
.ARM
.ALIGN 4
.DATA
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@
@ global pointers
@ uses of global pointers is the easiest way to implement this but it
@ consitutes common coupling which is not considered to be a particulary
@ good (or horrible) software engineering technique.
@
current_thread:     .word 0x0               @ pointer to thread that is running
ready_queue:        .word 0x0               @ queue of threads ready to run
suspended_queue:    .word 0x0               @ queue of suspended threads
@reg_if:             .word 0x04000202        @ interrupt flag register for GBA
reg_if:             .word 0x04000214        @ interrupt flag register for DS
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
dispatcher:
    @
    @ the dispatcher is probobly the most important part of any threading
    @ system. While the scheduler is responsible for determining who
    @ gets to execute when, the dispatcher is what actually executes
    @ and suspends the threads. The process of changing which thread
    @ is currently running on the system is referred to as "swapping."
    @ this code here will first "swap out" the thread that has been
    @ executing by making copies of all the registers and the state of
    @ the CPU in ram. Once that has been done, the thread can be resumed
    @ later as if it had never been interrupted in the first place.
    @ Next, the registers and CPU state of another thread, which have
    @ been stored in RAM since the last time that process was executing,
    @ are loaded back into the CPU and then a normal interrupt return
    @ is executed. The idea is that one thread will be interrupted by
    @ the hardware and when the interrupt handler finishes, another
    @ thread will continue executing. To the thread being resumed,
    @ things will be as if it was resuming from the hardware
    @ interupt that had caused it to be susspended in the first
    @ place, that is to say as if it had never been interrupted
    @ at all. This works semlessly because hardware interrupts are
    @ occuring all the time anyway and the CPU is designed
    @ to make them transparent to the running program.
    @
    @ first, get the pointer to the process discriptor for the
    @ currently running process. the "process discriptor" is
    @ the name of the memory block in RAM where a suspended
    @ process' CPU state is stored.
    ldr r0, current_thread          @ load pointer
    orrs r0, r0, r0                 @ check for NULL pointer
    addeq sp, sp, #24               @ discard pushes by BIOS if no pointer
    beq skip_swapout                @ skip swapout if no discriptor
    @ when an interrupt occurs, the registers r4 thru r11 are untoched
    @ by both the processor and the BIOS, as such they can simply be
    @ stored to the discriptor right away
    stmia r0!, {r4-r11}             @ save registers still in the CPU
    @ when a interrupt occurs, control is first directed to BIOS
    @ which pushes the registers r0 thru r3, r12 and r14
    @ before passing control to the gamepak interrupt handler, so
    @ just simply pop those values from the stack and store them in
    @ the thread discriptor. One thing to note, the value stored in
    @ r14 is really the interrupt return vector. i.e. although the
    @ value is stored in r14 it is really the value of r15 as far as
    @ the thread is concerned.
    ldmfd sp!, {r4-r9}              @ pop registers pushed by BIOS
    stmia r0!, {r4-r9}              @ save registers
    @
    @ use of registers in following code
    @
    @  r0   pointer offset in current_thread to where to store next register
    @  r1   return address to BIOS interupt handler
    @  r2   cpsr flags with CPU supervisor mode
    @  r3   cpsr flags with CPU system mode
    @  r4   cpsr flags with CPU IRQ mode
    @
    @ the processor state before the interrupt must be saved as well,
    @ it is available in the spsr_irq register
    mrs r5, spsr                    @ get the stored CPU state
    str r5, [r0], #4                @ store value for cpsr
    @
    @ the lr and sp registers are banked for each CPU mode, meaing
    @ that each mode has its own register which holds these values.
    @ since the CPU is in IRQ mode right now, accesses to lr and sp
    @ will result in accesses to lr_irq and sp_irq respectively. In
    @ order to fully susspend the thread, the values for lr and sp
    @ for both the supervisor and system/user modes must saved in
    @ addition to the spsr_svc register. The only way to accesses
    @ these registers is to put the CPU in the mode that those
    @ registers belong too and access them directly. The registers
    @ from the other CPU modes needn't be recorded because the
    @ those CPU modes should never be used in a GBA program.
    @
    @ warning! when changing the CPU mode be certain that you DO NOT
    @ touch bit 5 (0x20) because that is the ARM/THUMB state flag.
    @ changing it manually is not allowed.
    mrs r4, cpsr                    @ get cpu flags
    bic r4, r4, #0x1f               @ clear state code
    orr r2, r4, #0x13               @ supervisor mode
    orr r3, r4, #0x1f               @ system mode
    orr r4, r4, #0x12               @ irq mode
    @ store banked registers
    msr cpsr, r2                    @ put CPU in supervisor mode
    mrs r5, spsr                    @ store value of spsr_svc
    str r5, [r0], #4
    str lr, [r0], #4                @ store lr_svc
    str sp, [r0], #4                @ store sp_svc
    msr cpsr, r3                    @ put CPU in system mode
    str lr, [r0], #4                @ store lr
    str sp, [r0], #4                @ store sp
    msr cpsr, r4                    @ go back to IRQ mode for scheduling
skip_swapout:
    @ acknowledge all interrupts, which should only ever be the timer
    ldr r0, reg_if
    ldr r1, [r0]
    str r1, [r0]
    @
    @ execute the scheduler to determine who should execute next
    @
    @ this is an extremely simple round robin scheduler. just take the next
    @ thread to schedule off of a circular queue. this is probobly the
    @ most simple scheduler possible and probobly doesn't even diserve the
    @ name at all since it doesn't even include any thread priority.
    @
    ldr r0, ready_queue             @ get pointer off queue
    str r0, current_thread          @ update current_thread
    add r1, r0, #80                 @ get offset to next pointer
    ldr r1, [r1]                    @ get pointer to next item in queue
    str r1, ready_queue             @ store new pointer in queue
    add r0, r0, #80                 @ set pointer to end of registers block
    @ since the values for the various cpu modes were never calculated
    @ if no swap out took place, the values need to be calculated here
    mrs r4, cpsr                    @ get cpu flags
    bic r4, r4, #0x1f               @ clear state code
    orr r2, r4, #0x13               @ supervisor mode
    orr r3, r4, #0x1f               @ system mode
    orr r4, r4, #0x12               @ irq mode
    @
    @ to resume a thread, the process is basically reversed in direction
    @ (i.e. stores become loads) and reversed in order
    @
    @ restore banked registers
    msr cpsr, r3                    @ put CPU in system mode to get lr and sp
    ldr sp, [r0, #-4]!              @ restore sp
    ldr lr, [r0, #-4]!              @ restore lr
    msr cpsr, r2                    @ put CPU in supervisor mode
    ldr sp, [r0, #-4]!              @ restore sp_svc
    ldr lr, [r0, #-4]!              @ restore lr_svc
    ldr r5, [r0, #-4]!              @ restore spsr_svc
    msr spsr, r5
    msr cpsr, r4                    @ banked registers done, restore IRQ mode
    @ restore other registers
    ldr r5, [r0, #-4]!              @ get bits for CPSR
    msr spsr, r5                    @ cpsr = spsr_irq when interrupt returns
    ldmdb r0!, {r4-r9}              @ load values for r0-r3, r12, r14
    stmfd sp!, {r4-r9}              @ push values; BIOS will restore for us
    ldmdb r0!, {r4-r11}             @ restore general registers
    @ branch to thread    
    bx lr                           @ let BIOS resume from interrupt
.pool
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
