/*
 *  thread.c
 *
 *  Created by DarkPhantom.
 *
 *  Modified by samel on 28/02/08.
 *
 */

#include "main9.h"
#include "dispatch.h"

//brief: assign to each thread an ID
int threadid=0;

//------------------------------------------------------------------------------
thread* _create_thread(void *function_entry_point,
                     int user_stack_size, int svc_stack_size,
                     unsigned int function_param)
{
 thread *newthread;
 _disable_interrupt();
 if(user_stack_size == USE_DEFAULT_STACK_SIZE) user_stack_size = 128;
 if(svc_stack_size == USE_DEFAULT_STACK_SIZE) svc_stack_size = 64;
 newthread = malloc(sizeof(thread));
 if(!newthread) return NULL;
 bzero(newthread,sizeof(thread));
 newthread -> user_stack_size = malloc(sizeof(unsigned int)*user_stack_size);
 newthread -> svc_stack_size = malloc(sizeof(unsigned int)*svc_stack_size);
 if(!newthread -> user_stack_size || !newthread -> svc_stack_size)
 {
  if(newthread -> user_stack_size) free(newthread -> user_stack_size);
  if(newthread -> svc_stack_size) free(newthread -> svc_stack_size);
  free(newthread);
  return NULL;
 }
 newthread -> id = threadid++;
 newthread -> r0 = function_param;
 newthread -> pc = (unsigned int)function_entry_point;
 newthread -> lr = (unsigned int)_kill_self;
 newthread -> cpsr = 0x1f;
 newthread -> sp = (unsigned int)((newthread -> user_stack_size) + sizeof(unsigned int)*user_stack_size);
 newthread -> sp_svc = (unsigned int)((newthread -> svc_stack_size) + sizeof(unsigned int)*svc_stack_size);
 _enqueue_thread(&ready_queue, newthread);
 _enable_interrupt();
 return newthread;
}
//------------------------------------------------------------------------------
void _kill_thread(thread *t)
{
 _disable_interrupt();
 _dequeue_thread(t);
 free(t -> user_stack_size);
 free(t -> svc_stack_size);
 free(t);
 if( t == current_thread )
 {
  current_thread = NULL;
  _enable_interrupt();
  for(;;);
 }
 else
 {
  _enable_interrupt();
  for(;;);
 }
}
//------------------------------------------------------------------------------
void _resume_thread(thread *t)
{
 if(!t) return;
 _disable_interrupt();
 _dequeue_thread(t);
 _enqueue_thread(&ready_queue, t);
 _enable_interrupt();
}
//------------------------------------------------------------------------------
void _suspend_thread(thread *t)
{
 if(!t) return;
 _disable_interrupt();
 _dequeue_thread(t);
 _enqueue_thread(&suspended_queue, t);
 _enable_interrupt();
}
//------------------------------------------------------------------------------
void _enqueue_thread(thread **queue, thread *t)
{
 if(*queue == NULL)
 {
  *queue = t;
  t -> next = t;
  t -> prev = t;
 }
 else
 {
  t -> next = *queue;
  t -> prev = (*queue) -> prev;
  t -> next -> prev = t;
  t -> prev -> next = t;
 }
}
//------------------------------------------------------------------------------
void _append_queue(thread **dest, thread **src)
{
 thread *end_of_dest;
 thread *end_of_src;
 if(*dest)
 {
  if(!*src) return;
  end_of_dest = (*dest) -> prev;
  end_of_src = (*src) -> prev;
  (*dest) -> prev = end_of_src;
  (*src) -> prev = end_of_dest;
  end_of_dest -> next = *src;
  end_of_src -> next = *dest;
 }
 else *dest = *src;
 *src = NULL;
}
//------------------------------------------------------------------------------
void _dequeue_thread(thread *t)
{
 t -> next -> prev = t -> prev;
 t -> prev -> next = t -> next;
 if( t -> next == t) t -> next = t -> prev = NULL;
 if(t == ready_queue) ready_queue = t -> next;
 if(t == suspended_queue) suspended_queue = t -> next;
}
//------------------------------------------------------------------------------
void _kill_self(void)
{
 asm("ldr r0, =current_thread");
 asm("ldr r0, [r0]");
 asm("ldr lr, =_kill_thread");
}
//------------------------------------------------------------------------------
void _thread_init()
{
 irqInitHandler(dispatcher);
 irqEnable(IRQ_VBLANK);
}
