/*
 *  thread.h
 *
 *  Copyright (C) 2008 Giacometti Luca
 * 
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _INC_THREAD_
#define _INC_THREAD_

#define USE_DEFAULT_STACK_SIZE 0

typedef struct tagThread
{
 
 // DarkPhantom
 // the format of this structure may seem odd but it is designed to
 // facilitate the fastest possible swap-in/swap-out procedure.
 // the exact reasoning behind the ordering of the registers in this
 // structure is explained in, and should be selfevident from, the
 // assembly code in dispatch.s
 //
 // general purpose registers that are saved first
 
 // Samel
 //           0   4   8   12  16  20  24   28  register offset
 unsigned int r4, r5, r6, r7, r8, r9, r10, r11;
 
 // DarkPhantom
 // registers pushed by the BIOS that are poped and then saved
 // note: pc is really the value or lr_irq but is restored to pc
 // upon the interrupt return
 
 // Samel
 //           32  36  40  44  48   52  register offset
 unsigned int r0, r1, r2, r3, r12, pc;
 // DarkPhantom
 // value from spsr_irq
 // Samel
 //           56  register offset
 unsigned int cpsr;
 // DarkPhantom
 // svc mode banked registers
 // Samel
 //           60        64      68  register offset
 unsigned int spsr_svc, lr_svc, sp_svc;
 // DarkPhantom
 // system mode banked registers
 // Samel
 //           72  76  register offset
 unsigned int lr, sp;
 
 // DarkPhantom
 // queued list functions
 struct tagThread *next;
 struct tagThread *prev;
 
 // DarkPhantom
 // dynamic stack pointers
 void *user_stack_size;
 void *svc_stack_size;
 
 int id;
 
} thread;

/*
 brief: create a new thread
 function_entry_point: pointer to thread function
 user_stack_size: user-mode stack size in unsigned int
 svc_stack_size: system-mode stack size in unsigned int
 function_param: function parameter
 return: pointer to thread structure
 */
thread* _create_thread(void *function_entry_point,
                     int user_stack_size, int svc_stack_size, unsigned int function_param);

/*
 brief: kill a running thread
 t: pointer to the thread structure
 */
void _kill_thread(thread *t);

/*
 brief: resume a suspended thread
 t: pointer to the thread structure
 */
void _resume_thread(thread *t);

/*
 brief: suspend a thread (put it in the suspendqueue)
 t: pointer to the thread structure
 */
void _suspend_thread(thread *t);

/*
 brief: add a thread to the selected queue
 queue: the queue where you want to attach the thread
 t: pointer to the thread structure
 */
void _enqueue_thread(thread **queue, thread *t);

/*
 brief: merge two thread queue
 dest: the queue where you want to attach the src queue
 dest: the queue you want to attach
 */
void _append_queue(thread **dest, thread **src);

/*
 brief: remove the thread from any queue
 t: pointer to the thread structure
 */
void _dequeue_thread(thread *t);

/*
 brief: function that replace the return C/C++ stantment
 */
void _kill_self(void);

/*
 brief: initialize the dispatcher and let thread start
 */
void _thread_init();

#endif
